
#include <..\AutoitObject.au3>
#include <String.au3>
#include <Crypt.au3>

Global Const $__TCPClass_BaseClientPropertiesNbr = 4
Global $__TCPServer_Sockets[1][$__TCPClass_BaseClientPropertiesNbr] ; ONLY ONE SERVER PER SCRIPT!
Global $__TCPServer_hKey = 0

Global $__TCPClass_oMyErr = ObjEvent("AutoIt.Error", "_TCPClass_ErrFunc")

; ##############################################################
#cs
	File: TCPServer
#ce

#cs
Class: _Class_TCPServer
	Object that simplifiy the creation of Encrypted, multi-clients TCP Server.
	
	There can be *only ONE server* per script.

Property: ip
	IP Address to listen

Property: port
	Port number to listen

Property: max_recv_len
	max # of characters to receive (TCPRecv)

Property: Timeout
	Receiving timeout (in ms).
	
	When this timeout is reached, the buffer is fluched, and the data that was
	being transfered is lost.
	
	See <Callback_Timedout>

Property: NbrClients
	Number of clients currently connected to the server

Property: Callback_NewClient
	Function called when a new client connects to the server.
	
	The function must be as this prototype:
	
	>_Func($iSocket, $sIP)
	
	$iSocket - Socket ID of the client
	$sIP - IP Adress of the client

Property: Callback_LostClient
	Function called when a client disconnects from the server.
	
	The function must be as this prototype:
	
	>_Func($iSocket, $sIP)
	
	$iSocket - Socket ID of the client
	$sIP - IP Adress of the client

Property: Callback_Recv
	Function that is called when data (Binary/String) is received from a client
	
	The function must be as this prototype:
	
	>_Func($iSocket, $sIP, $s_bData)
	
	$iSocket - Socket ID of the client
	$sIP - IP Adress of the client
	$s_bData - Received data (Binary/String)

Property: Callback_Receiving
	Function that is called when being receiving data from a client
	
	The function must be as this prototype:
	
	>_Func($iSocket, $sIP, $BufferLenght)
	
	$iSocket - Socket ID of the client
	$sIP - IP Adress of the client
	$BufferLenght - Lenght of the client's receiving buffer (Not equal to the final lenght/size
		of the received data, it's just an indication)

Property: Callback_Timedout
	Function that is called when a client is timed out
	
	The function must be as this prototype:
	
	>_Func($iSocket, $sIP, $BufferLenght_LostData)
	
	$iSocket - Socket ID of the client
	$sIP - IP Adress of the client
	$BufferLenght_LostData - Lenght of the buffer when timed out, corresponds to the amount of
		lost data (because buffer is flushed when time out occures)
#ce

#cs
Constructor: _Class_TCPServer
	Class constructor

Parameters:
	$ip - Value to assign to <ip>
	$port - Value to assign to <port>
	$max_clients - Maximum number of clients at the same time
	$Ext_Info_Nbr - Number of extended data that can be associated with each client (see <ClientPropertySet> & <ClientPropertyGet>)
	$encrypt_pwd - Data encryption password. Must be the same as in the clients.
#ce
Func _Class_TCPServer($ip = "", $port = -1, $max_clients = 50, $Ext_Info_Nbr = 0, $encrypt_pwd = "topsecret")
	Local $o = _AutoItObject_Class()
	_Crypt_Startup()
	$__TCPServer_hKey = _Crypt_DeriveKey($encrypt_pwd, $CALG_RC4)
	; ---
	With $o
		; ### Properties
		.AddProperty("ip", $ELSCOPE_PUBLIC, $ip)
		.AddProperty("port", $ELSCOPE_PUBLIC, $port)
		.AddProperty("max_recv_len", $ELSCOPE_PUBLIC, 4096)
		.AddProperty("Callback_NewClient", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_Recv", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_Receiving", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_LostClient", $ELSCOPE_PUBLIC, "")
		.AddProperty("Callback_Timedout", $ELSCOPE_PUBLIC, "")
		.AddProperty("NbrClients", $ELSCOPE_PUBLIC, 0)
		.AddProperty("Timeout", $ELSCOPE_PUBLIC, 10000)
		; --- Privates
		.AddProperty("socket", $ELSCOPE_PRIVATE, -1)
		.AddProperty("max_clients", $ELSCOPE_PRIVATE, $max_clients)
		.AddProperty("ext_info_nbr", $ELSCOPE_PRIVATE, $Ext_Info_Nbr)
		; ### Methods
		.AddMethod("Startup", "__Class_TCPServer_Startup")
		.AddMethod("Shutdown", "__Class_TCPServer_Shutdown")
		.AddMethod("Process", "__Class_TCPServer_Process")
		.AddMethod("Send", "__Class_TCPServer_Send")
		.AddMethod("Broadcast", "__Class_TCPServer_Broadcast")
		.AddMethod("Disconnect", "__Class_TCPServer_Disconnect")
		.AddMethod("DisconnectAll", "__Class_TCPServer_DisconnectAll")
		.AddMethod("SocketIDList", "__Class_TCPServer_SocketIDList")
		.AddMethod("SocketID2IP", "__Class_TCPServer_SocketID2IP")
		.AddMethod("IP2SocketID", "__Class_TCPServer_IP2SocketID")
		.AddMethod("SocketID2hSocket", "__Class_TCPServer_SocketID2hSocket")
		.AddMethod("ClientPropertySet", "__Class_TCPServer_ClientPropertySet")
		.AddMethod("ClientPropertyGet", "__Class_TCPServer_ClientPropertyGet")
		; --- Privates
		.AddMethod("Cry", "__Class_TCPClient_Crypt", True)
		.AddMethod("DCry", "__Class_TCPClient_UnCrypt", True)
		.AddMethod("CheckBuffer", "__Class_TCPClient_CheckBuffer", True)
		.AddMethod("SocketToIP", "__Class_TCPClient_SocketToIP", True)
		; --- Destructor
		.AddDestructor("__Class_TCPServer_Destructor")
	EndWith
	Return $o.object
EndFunc   ;==>_Class_TCPServer

; ##############################################################

#cs
Method: Startup
	Starts the server.

Parameters:
	$cb_NewClient - Value to assign to <Callback_NewClient>
	$cb_LostClient - Value to assign to <Callback_LostClient>
	$cb_Recv - Value to assign to <Callback_Recv>
	$cb_Receiving - Value to assign to <Callback_Receiving>
	$cb_Timedout - Value to assign to <Callback_Timedout>

Return value:
	Success - Returns main socket identifier.
	Failure - Returns -1 or 0 and set @error to:
 
	- 1 IPAddr is incorrect.
	- 2 port is incorrect.
	
	windows API WSAGetError return value (see MSDN).
#ce
Func __Class_TCPServer_Startup($oSelf, $cb_NewClient = "", $cb_LostClient = "", _
										$cb_Recv = "", $cb_Receiving = "", _
										$cb_Timedout = "")
	#forceref $oSelf
	; ---
	If $oSelf.socket <> -1 Then $oSelf.Shutdown()
	; ---
	Local $socket = TCPListen($oSelf.ip, $oSelf.port)
	If $socket <= 0 Or @error Then Return SetError(@error, 0, $socket)
	; ---
	$oSelf.Callback_NewClient = $cb_NewClient ; _Func($socket, $ip)
	$oSelf.Callback_Recv = $cb_Recv ; _Func($socket, $ip, $sData)
	$oSelf.Callback_Receiving = $cb_Receiving ; _Func($socket, $ip, $bufferLenght)
	$oSelf.Callback_LostClient = $cb_LostClient ; _Func($socket, $ip)
	$oSelf.Callback_Timedout = $cb_Timedout ; _Func($socket, $ip, $bufferLenght_lostdata)
	; ---
	ReDim $__TCPServer_Sockets[$oSelf.max_clients][$__TCPClass_BaseClientPropertiesNbr + $oSelf.ext_info_nbr]
	For $i = 0 To $oSelf.max_clients - 1
		_TCPServer_InitSocketArray($i, $oSelf.ext_info_nbr)
	Next
	; ---
	$oSelf.socket = $socket
	Return $socket
EndFunc   ;==>__Class_TCPServer_Startup

#cs
Method: Shutdown
	Disconnect all clients, and shutdown the server

Return value:
	Success - Returns 1. 
	Failure - Returns 0 and set @error = windows API WSAGetError return value (see MSDN).
#ce
Func __Class_TCPServer_Shutdown($oSelf)
	#forceref $oSelf
	; ---
	$oSelf.DisconnectAll()
	; ---
	Local $ret = TCPCloseSocket($oSelf.socket)
	Local $err = @error
	$oSelf.socket = -1
	Return SetError($err, 0, $ret)
EndFunc   ;==>__Class_TCPServer_Shutdown

#cs
Method: Process
	Function that calls all the server callbacks when events occures (New client, Receiving...)

Remark:
	This function must be called in loop, or in AdlibRegister
#ce
Func __Class_TCPServer_Process($oSelf)
	#forceref $oSelf
	; ---
	If $oSelf.socket = -1 Then Return SetError(1, 0, 0)
	; ---
	Local $new_socket = TCPAccept($oSelf.socket)
	If $new_socket <> -1 Then
		Local $stored = 0
		For $i = 0 To $oSelf.max_clients - 1
			If $__TCPServer_Sockets[$i][0] = -1 Then
				$stored = 1
				$oSelf.NbrClients = $oSelf.NbrClients + 1
				; ---
				$__TCPServer_Sockets[$i][0] = $new_socket
				$__TCPServer_Sockets[$i][1] = $oSelf.SocketToIP($new_socket)
				; ---
				If $oSelf.Callback_NewClient <> "" Then
					;Call($oSelf.Callback_NewClient, $__TCPServer_Sockets[$i][0], $__TCPServer_Sockets[$i][1])
					Call($oSelf.Callback_NewClient, $i, $__TCPServer_Sockets[$i][1])
				EndIf
				; ---
				ExitLoop
			EndIf
		Next
		If Not $stored Then TCPCloseSocket($new_socket)
	EndIf
	; ---
	Local $sRecv, $ret
	For $i = 0 To $oSelf.max_clients - 1
		If $__TCPServer_Sockets[$i][0] = -1 Then ContinueLoop
		; ---
		$sRecv = TCPRecv($__TCPServer_Sockets[$i][0], $oSelf.max_recv_len)
		If @error Then
			; If it was previously connected (a valide socket)
			If $__TCPServer_Sockets[$i][0] <> -1 Then
				If $oSelf.Callback_LostClient <> "" Then
					;Call($oSelf.Callback_LostClient, $__TCPServer_Sockets[$i][0], $__TCPServer_Sockets[$i][1])
					Call($oSelf.Callback_LostClient, $i, $__TCPServer_Sockets[$i][1])
				EndIf
				;$oSelf.Disconnect($__TCPServer_Sockets[$i][0])
				$oSelf.Disconnect($i)
				$oSelf.NbrClients = $oSelf.NbrClients - 1
				_TCPServer_InitSocketArray($i, $oSelf.ext_info_nbr)
			EndIf
			ContinueLoop
		EndIf
		; ---
		If $sRecv Then
			$__TCPServer_Sockets[$i][3] = TimerInit()
			$__TCPServer_Sockets[$i][2] &= String($sRecv)
			; call receiving callback
			If $oSelf.Callback_Receiving <> "" Then
				;Call($oSelf.Callback_Receiving, $__TCPServer_Sockets[$i][0], $__TCPServer_Sockets[$i][1], BinaryLen($__TCPServer_Sockets[$i][2]))
				Call($oSelf.Callback_Receiving, $i, $__TCPServer_Sockets[$i][1], BinaryLen($__TCPServer_Sockets[$i][2]))
			EndIf
		EndIf
		; Timeout check
		If $__TCPServer_Sockets[$i][2] Then
			If TimerDiff($__TCPServer_Sockets[$i][3]) >= $oSelf.Timeout Then
				; Call timedout callbakc
				If $oSelf.Callback_Timedout <> "" Then
					;Call($oSelf.Callback_Timedout, $__TCPServer_Sockets[$i][0], $__TCPServer_Sockets[$i][1], BinaryLen($__TCPServer_Sockets[$i][2]))
					Call($oSelf.Callback_Timedout, $i, $__TCPServer_Sockets[$i][1], BinaryLen($__TCPServer_Sockets[$i][2]))
				EndIf
				; Flush the buffer
				$__TCPServer_Sockets[$i][2] = ""
			EndIf
		EndIf
		; ---
		If $oSelf.CheckBuffer($i) Then
			Local $ret = _StringBetween($__TCPServer_Sockets[$i][2], Chr(2), Chr(3))
			For $elem In $ret
				$ret = $oSelf.DCry($elem)
				If $oSelf.Callback_Recv <> "" Then
					;Call($oSelf.Callback_Recv, $__TCPServer_Sockets[$i][0], $__TCPServer_Sockets[$i][1], $ret)
					Call($oSelf.Callback_Recv, $i, $__TCPServer_Sockets[$i][1], $ret)
				EndIf
			Next
			$__TCPServer_Sockets[$i][2] = ""
		EndIf
	Next
EndFunc   ;==>__Class_TCPServer_Process

#cs
Method: Send
	Send data (Binary/String) to a client

Parameters:
	$iSocket - Socket ID of the client
	$s_bData - The data to be sent (Binary/String)

Return value:
	Success - Returns number of bytes sent to the server. 
	Failure - Returns 0 and set @error = windows API WSAGetError return value (see MSDN).
#ce
Func __Class_TCPServer_Send($oSelf, $iSocket, $s_bData)
	#forceref $oSelf
	; ---
	If $oSelf.socket = -1 Then Return SetError(1, 0, 0)
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(2, 0, 0)
	; ---
	If IsString($s_bData) Then
		$s_bData = StringReplace($s_bData, "'", "''")
	EndIf
	; ---
	Return SetError(@error, @extended, TCPSend($__TCPServer_Sockets[$iSocket][0], $oSelf.Cry($s_bData)))
EndFunc   ;==>__Class_TCPServer_Send

#cs
Method: Broadcast
	Send data to all connected clients

Parameters:
	$s_bData - Data to be sent (Binary/String)

Return value:
	Success - Number of clients that data was sent to them successfully,
		@extended = Percentage of succes (Return value / Total number of clients)
	Failure - 0 and
		
		@error = 1 -> Server not started
		@error = 2 -> No any client received data
#ce
Func __Class_TCPServer_Broadcast($oSelf, $s_bData)
	#forceref $oSelf
	; ---
	If $oSelf.socket = -1 Then Return SetError(1, 0, 0)
	; ---
	If IsString($s_bData) Then
		$s_bData = StringReplace($s_bData, "'", "''")
	EndIf
	; ---
	Local $nbr = 0
	For $i = 0 To $oSelf.max_clients - 1
		If $__TCPServer_Sockets[$i][0] <> -1 Then
			TCPSend($__TCPServer_Sockets[$i][0], $oSelf.Cry($s_bData))
			If Not @error Then $nbr += 1
		EndIf
	Next
	; ---
	Local $err = 0
	If $nbr = 0 Then $err = 2
	Return SetError($err, $nbr / $oSelf.NbrClients, $nbr)
EndFunc   ;==>__Class_TCPServer_Broadcast

#cs
Method: Disconnect
	Disconnect a specified client

Parameters:
	$iSocket - Socket ID of the client to disconnect
#ce
Func __Class_TCPServer_Disconnect($oSelf, $iSocket)
	#forceref $oSelf
	; ---
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(1, 0, 0)
	; ---
	; On initialise pas ici le slot dans le array des sockets
	; pour que la fonction callback de deconnection soit appel correctement.
	Return TCPCloseSocket($__TCPServer_Sockets[$iSocket][0])
EndFunc   ;==>__Class_TCPServer_Disconnect

#cs
Method: DisconnectAll
	Disconnect all connected clients.
#ce
Func __Class_TCPServer_DisconnectAll($oSelf)
	#forceref $oSelf
	; ---
	For $i = 0 To $oSelf.max_clients - 1
		TCPCloseSocket($__TCPServer_Sockets[$i][0])
	Next
EndFunc   ;==>__Class_TCPServer_DisconnectAll

#cs
Method: SocketIDList
	Return a 1-based array of all currently connected client's socket ID

Return value:
	1-based array of connected sockets ID, $elem[0] = Number of sockets
#ce
Func __Class_TCPServer_SocketIDList($oSelf)
	#forceref $oSelf
	; ---
	Local $ret[1] = [0]
	For $i = 0 To $oSelf.max_clients - 1
		If $__TCPServer_Sockets[$i][0] <> -1 Then
			Redim $ret[$ret[0] + 2]
			;$ret[$ret[0] + 1] = $__TCPServer_Sockets[$i][0]
			$ret[$ret[0] + 1] = $i
			$ret[0] += 1
		EndIf
	Next
	; ---
	Return $ret
EndFunc

#cs
Method: SocketID2IP
	Return the IP of a given client's socket

Parameters:
	$iSocket - Socket ID of a client
#ce
Func __Class_TCPServer_SocketID2IP($oSelf, $iSocket)
	#forceref $oSelf
	; ---
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(1, 0, 0)
	; ---
	;For $i = 0 To $oSelf.max_clients - 1
		;If $__TCPServer_Sockets[$i][0] = $iSocket Then Return $__TCPServer_Sockets[$i][1]
		If $__TCPServer_Sockets[$iSocket][0] <> -1 Then Return $__TCPServer_Sockets[$iSocket][1]
	;Next
	; ---
	Return SetError(2, 0, 0)
EndFunc

#cs
Method: IP2SocketID
	Return the socket ID of a given client's IP

Parameters:
	$sIP - IP of a client
#ce
Func __Class_TCPServer_IP2SocketID($oSelf, $sIP)
	#forceref $oSelf
	; ---
	For $i = 0 To $oSelf.max_clients - 1
		;If $__TCPServer_Sockets[$i][1] = $sIP Then Return $__TCPServer_Sockets[$i][0]
		If $__TCPServer_Sockets[$i][1] = $sIP Then Return $i
	Next
	; ---
	Return SetError(1, 0, 0)
EndFunc

#cs
Method: SocketID2hSocket
	Return the Socket Handle of a client's Socket ID

Parameters:
	$iSocket - Socket ID of a client
#ce
Func _Class_TCPServer_SocketID2hSocket($oSelf, $iSocket)
	#forceref $oSelf
	; ---
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(1, 0, 0)
	; ---
	If $__TCPServer_Sockets[$iSocket][0] <> -1 Then Return $__TCPServer_Sockets[$iSocket][0]
	; ---
	Return SetError(2, 0, -1)
EndFunc

#cs
Method: ClientPropertySet
	Set an extended property of a client to a specified value

Parameters:
	$hSocket - Socket ID of the client
	$iIndex - Index of the property we want to modifie
	$Data - Data to assign to the property

Return value:
	
#ce
Func __Class_TCPServer_ClientPropertySet($oSelf, $iSocket, $iIndex, $Data)
	#forceref $oSelf
	; ---
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(1, 0, 0)
	If $iIndex >= $oSelf.ext_info_nbr Then Return SetError(2, 0, 0)
	; ---
	;For $i = 0 To $oSelf.max_clients - 1
	;	If $hSocket = $__TCPServer_Sockets[$i][0] Then
			$__TCPServer_Sockets[$iSocket][$iIndex + $__TCPClass_BaseClientPropertiesNbr] = $Data
	;	EndIf
	;Next
	; ---
	Return SetError(0, 0, 1)
EndFunc

#cs
Method: ClientPropertyGet
	Get the value of a client's extended property

Parameters:
	$hSocket - Socket ID of the client
	$iIndex - Index of the property we want to get

Return value:
	
#ce
Func __Class_TCPServer_ClientPropertyGet($oSelf, $iSocket, $iIndex)
	#forceref $oSelf
	; ---
	If $iSocket >= $oSelf.max_clients Or $iSocket < 0 Then Return SetError(1, 0, 0)
	If $iIndex >= $oSelf.ext_info_nbr Then Return SetError(2, 0, 0)
	; ---
	;For $i = 0 To $oSelf.max_clients - 1
	;	If $hSocket = $__TCPServer_Sockets[$i][0] Then
			Return $__TCPServer_Sockets[$iSocket][$iIndex + $__TCPClass_BaseClientPropertiesNbr]
	;	EndIf
	;Next
	; ---
	Return SetError(0, 0, 1)
EndFunc

; ##############################################################
; ### Internals ################################################
; ##############################################################

Func __Class_TCPClient_Crypt($oSelf, $data)
	#forceref $oSelf
	; ---
	Local $ret
	; ---
	If Not IsBinary($data) Then $data = String($data)
	If IsBinary($data) Then
		$ret = Chr(2) & String(_Crypt_EncryptData($data, $__TCPServer_hKey, $CALG_USERKEY)) & Chr(3)
	Else
		$ret = Chr(2) & String(StringTrimLeft(_Crypt_EncryptData($data, $__TCPServer_hKey, $CALG_USERKEY), 2)) & Chr(3)
	EndIf
	; ---
	;ConsoleWrite("Crypt: " & $data & " (" & VarGetType($data) & ") => " & $ret & " (" & VarGetType($ret) & ")" & @CRLF)
	Return $ret
EndFunc   ;==>__Class_TCPClient_Crypt

Func __Class_TCPClient_UnCrypt($oSelf, $data)
	#forceref $oSelf
	; ---
	Local $ret
	; ---
	If StringLeft($data, 2) = "0x" Then
		$ret = _Crypt_DecryptData($data, $__TCPServer_hKey, $CALG_USERKEY)
	Else
		$ret = BinaryToString(_Crypt_DecryptData(Binary("0x" & $data), $__TCPServer_hKey, $CALG_USERKEY))
	EndIf
	; ---
	;ConsoleWrite("DCry: " & $data & " (" & VarGetType($data) & ") => " & $ret & " (" & VarGetType($ret) & ")" & @CRLF)
	Return $ret
EndFunc   ;==>__Class_TCPClient_UnCrypt

Func __Class_TCPClient_CheckBuffer($oSelf, $i)
	#forceref $oSelf
	; ---
	If StringLeft($__TCPServer_Sockets[$i][2], 1) = Chr(2) And StringRight($__TCPServer_Sockets[$i][2], 1) = Chr(3) Then
		Return 1
	Else
		Return 0
	EndIf
EndFunc   ;==>__Class_TCPClient_CheckBuffer

; ##############################################################

Func __Class_TCPServer_Destructor($oSelf)
	#forceref $oSelf
	; ---
	_Crypt_DestroyKey($__TCPServer_hKey)
	_Crypt_Shutdown()
	$oSelf.Shutdown()
EndFunc   ;==>__Class_TCPServer_Destructor

; ##############################################################

Func __Class_TCPClient_SocketToIP($oSelf, $SHOCKET)
	#forceref $oSelf
	; ---
	Local $sockaddr, $aRet
	$sockaddr = DllStructCreate("short;ushort;uint;char[8]")
	$aRet = DllCall("Ws2_32.dll", "int", "getpeername", "int", $SHOCKET, _
			"ptr", DllStructGetPtr($sockaddr), "int*", DllStructGetSize($sockaddr))
	If Not @error And $aRet[0] = 0 Then
		$aRet = DllCall("Ws2_32.dll", "str", "inet_ntoa", "int", DllStructGetData($sockaddr, 3))
		If Not @error Then $aRet = $aRet[0]
	Else
		$aRet = 0
	EndIf
	$sockaddr = 0
	Return $aRet
EndFunc   ;==>__Class_TCPClient_SocketToIP

; ##############################################################

Func _TCPServer_InitSocketArray($i, $extInfoNbr = 0)
	$__TCPServer_Sockets[$i][0] = -1 ; socket
	$__TCPServer_Sockets[$i][1] = "" ; ip
	$__TCPServer_Sockets[$i][2] = "" ; recv buffer
	$__TCPServer_Sockets[$i][3] = -1 ; recv timer (not implemented yet)
	; ---
	If $extInfoNbr > 0 Then
		For $y = $__TCPClass_BaseClientPropertiesNbr To $__TCPClass_BaseClientPropertiesNbr + $extInfoNbr
			$__TCPServer_Sockets[$i][$y] = 0
		Next
	EndIf
EndFunc   ;==>_TCPServer_InitSocketArray

; ##############################################################

Func _TCPClass_ErrFunc()
    ConsoleWrite("COM Error, ScriptLine(" & $__TCPClass_oMyErr.scriptline & ") : Number 0x" & Hex($__TCPClass_oMyErr.number, 8) & " - " & $__TCPClass_oMyErr.windescription & @CRLF)
EndFunc    ;==>_ErrFunc
